@@ -3,8 +3,9 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2008-2016, University of Amsterdam
+    Copyright (c)  2008-2020, University of Amsterdam
                               VU University Amsterdam
+                              CWI Amsterdam
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -41,11 +42,15 @@
             aggregate_all/4,            % +Templ, +Discrim, :Goal, -Result
             free_variables/4            % :Generator, :Template, +Vars0, -Vars
           ]).
-:- use_module(library(ordsets)).
-:- use_module(library(pairs)).
-:- use_module(library(error)).
-:- use_module(library(lists)).
-:- use_module(library(apply)).
+:- autoload(library(apply),[maplist/4,maplist/5]).
+:- autoload(library(error),
+	    [instantiation_error/1,type_error/2,domain_error/2]).
+:- autoload(library(lists),
+	    [append/3,member/2,sum_list/2,max_list/2,min_list/2]).
+:- autoload(library(ordsets),[ord_subtract/3,ord_intersection/3]).
+:- autoload(library(pairs),[pairs_values/2]).
+
+:- set_prolog_flag(generate_debug_info, false).
 
 :- meta_predicate
     foreach(0,0),
@@ -58,23 +63,24 @@
 
 This library provides aggregating operators  over   the  solutions  of a
 predicate. The operations are a generalisation   of the bagof/3, setof/3
-and findall/3 built-in predicates. The   defined  aggregation operations
-are counting, computing the sum, minimum,   maximum,  a bag of solutions
-and a set of solutions. We first   give  a simple example, computing the
-country with the smallest area:
+and findall/3 built-in predicates.  Aggregations   that  can be computed
+incrementally avoid findall/3 and run in   constant  memory. The defined
+aggregation  operations  are  counting,  computing   the  sum,  minimum,
+maximum, a bag of solutions and a  set   of  solutions.  We first give a
+simple example, computing the country with the smallest area:
 
-==
+```
 smallest_country(Name, Area) :-
-        aggregate(min(A, N), country(N, A), min(Area, Name)).
-==
+    aggregate(min(A, N), country(N, A), min(Area, Name)).
+```
 
 There are four aggregation predicates (aggregate/3, aggregate/4, aggregate_all/3 and aggregate/4), distinguished on two properties.
 
     $ aggregate vs. aggregate_all :
     The aggregate predicates use setof/3 (aggregate/4) or bagof/3
     (aggregate/3), dealing with existential qualified variables
-    (Var^Goal) and providing multiple solutions for the remaining free
-    variables in Goal. The aggregate_all/3 predicate uses findall/3,
+    (`Var^Goal`) and providing multiple solutions for the remaining free
+    variables in `Goal`. The aggregate_all/3 predicate uses findall/3,
     implicitly qualifying all free variables and providing exactly one
     solution, while aggregate_all/4 uses sort/2 over solutions that
     Discriminator (see below) generated using findall/3.
@@ -161,12 +167,15 @@ aggregate(Template, Discriminator, Goal0, Result) :-
 
 %!  aggregate_all(+Template, :Goal, -Result) is semidet.
 %
-%   Aggregate  bindings  in  Goal   according    to   Template.  The
-%   aggregate_all/3 version performs findall/3 on   Goal.  Note that
-%   this predicate fails if Template contains one or more of min(X),
-%   max(X),  min(X,Witness)  or  max(X,Witness)  and   Goal  has  no
-%   solutions, i.e., the minumum and  maximum   of  an  empty set is
-%   undefined.
+%   Aggregate   bindings   in   Goal   according    to   Template.   The
+%   aggregate_all/3 version performs findall/3 on   Goal. Note that this
+%   predicate fails if Template contains one  or more of min(X), max(X),
+%   min(X,Witness) or max(X,Witness) and Goal   has  no solutions, i.e.,
+%   the minimum and maximum of an empty set is undefined.
+%
+%   The Template values `count`, sum(X),   max(X),  min(X), max(X,W) and
+%   min(X,W) are processed incrementally rather than using findall/3 and
+%   run in constant memory.
 
 aggregate_all(Var, _, _) :-
     var(Var),
@@ -537,47 +546,61 @@ state1(_,   X, X, _).
 
 %!  foreach(:Generator, :Goal)
 %
-%   True if conjunction of results is   true. Unlike forall/2, which
-%   runs a failure-driven loop that proves Goal for each solution of
-%   Generator, foreach/2 creates a conjunction.   Each member of the
-%   conjunction is a copy of  Goal,   where  the variables it shares
-%   with Generator are filled with the values from the corresponding
-%   solution.
+%   True when the conjunction  of  _instances_   of  Goal  created  from
+%   solutions for Generator is true. Except for term copying, this could
+%   be implemented as below.
 %
-%   The implementation executes forall/2 if   Goal  does not contain
-%   any variables that are not shared with Generator.
+%   ```
+%   foreach(Generator, Goal) :-
+%       findall(Goal, Generator, Goals),
+%       maplist(call, Goals).
+%   ```
 %
-%   Here is an example:
+%   The actual implementation uses findall/3 on  a template created from
+%   the variables _shared_ between Generator  and Goal. Subsequently, it
+%   uses every instance of this template  to instantiate Goal, call Goal
+%   and undo _only_ the instantiation of   the  template and _not_ other
+%   instantiations created by running Goal.  Here is an example:
 %
-%   ==
+%   ```
 %   ?- foreach(between(1,4,X), dif(X,Y)), Y = 5.
 %   Y = 5.
 %   ?- foreach(between(1,4,X), dif(X,Y)), Y = 3.
 %   false.
-%   ==
+%   ```
 %
-%   @bug    Goal is copied repeatedly, which may cause problems if
-%           attributed variables are involved.
+%   The  predicate  foreach/2  is   mostly    used   if   Goal  performs
+%   backtrackable destructive assignment on  terms. Attributed variables
+%   (underlying constraints) are  an  example.   Another  example  of  a
+%   backtrackable data structure is in   library(hashtable).  If we care
+%   only about the side effects  (I/O,   dynamic  database, etc.) or the
+%   truth value of Goal,  forall/2 is  a faster and simpler alternative.
+%   If Goal instantiates its arguments it  is   will  often  fail as the
+%   argument cannot be instantiated to multiple   values. It is possible
+%   to incrementally _grow_ an argument:
+%
+%   ```
+%   ?- foreach(between(1,4,X), member(X, L)).
+%   L = [1,2,3,4|_].
+%   ```
+%
+%   Note that SWI-Prolog up to  version   8.3.4  created  copies of Goal
+%   using copy_term/2 for each iteration.
 
 foreach(Generator, Goal) :-
     term_variables(Generator, GenVars0), sort(GenVars0, GenVars),
     term_variables(Goal, GoalVars0), sort(GoalVars0, GoalVars),
-    ord_subtract(GoalVars, GenVars, SharedGoalVars),
-    (   SharedGoalVars == []
-    ->  \+ (Generator, \+Goal)      % = forall(Generator, Goal)
-    ;   ord_intersection(GenVars, GoalVars, SharedVars),
-        Templ =.. [v|SharedVars],
-        SharedTempl =.. [v|SharedGoalVars],
-        findall(Templ, Generator, List),
-        prove_list(List, Templ, SharedTempl, Goal)
-    ).
+    ord_intersection(GenVars, GoalVars, SharedVars),
+    Templ =.. [v|SharedVars],
+    findall(Templ, Generator, List),
+    prove_list(List, Templ, Goal).
 
-prove_list([], _, _, _).
-prove_list([H|T], Templ, SharedTempl, Goal) :-
-    copy_term(Templ+SharedTempl+Goal,
-              H+SharedTempl+Copy),
-    Copy,
-    prove_list(T, Templ, SharedTempl, Goal).
+prove_list([], _, _).
+prove_list([H|T], Templ, Goal) :-
+    Templ = H,
+    call(Goal),
+    '$unbind_template'(Templ),
+    prove_list(T, Templ, Goal).
 
 
 %!  free_variables(:Generator, +Template, +VarList0, -VarList) is det.
