@@ -3,8 +3,10 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  2005-2016, University of Amsterdam
+    Copyright (c)  2005-2021, University of Amsterdam
                               VU University Amsterdam
+                              CWI, Amsterdam
+                              SWI-Prolog Solutions b.v.
     All rights reserved.
 
     Redistribution and use in source and binary forms, with or without
@@ -35,16 +37,19 @@
 
 :- module(prolog_clause,
           [ clause_info/4,              % +ClauseRef, -File, -TermPos, -VarNames
+            clause_info/5,              % +ClauseRef, -File, -TermPos, -VarNames,
+                                        % +Options
             initialization_layout/4,    % +SourceLoc, +Goal, -Term, -TermPos
             predicate_name/2,           % +Head, -Name
             clause_name/2               % +ClauseRef, -Name
           ]).
-:- use_module(library(lists), [append/3]).
-:- use_module(library(occurs), [sub_term/2]).
-:- use_module(library(debug)).
-:- use_module(library(option)).
-:- use_module(library(listing)).
-:- use_module(library(prolog_source)).
+:- autoload(library(debug),[debugging/1,debug/3]).
+:- autoload(library(listing),[portray_clause/1]).
+:- autoload(library(lists),[append/3]).
+:- autoload(library(occurs),[sub_term/2]).
+:- autoload(library(option),[option/3]).
+:- autoload(library(prolog_source),[read_source_term_at_location/3]).
+
 
 :- public                               % called from library(trace/clause)
     unify_term/2,
@@ -58,7 +63,9 @@
     open_source/2.                  % +Input, -Stream
 
 :- predicate_options(prolog_clause:clause_info/5, 5,
-                     [ variable_names(-list)
+                     [ head(-any),
+                       body(-any),
+                       variable_names(-list)
                      ]).
 
 /** <module> Get detailed source-information about a clause
@@ -107,6 +114,8 @@ clause_info(ClauseRef, File, TermPos, NameOffset, Options) :-
     clause_property(ClauseRef, file(File)),
     File \== user,                  % loaded using ?- [user].
     '$clause'(Head0, Body, ClauseRef, VarOffset),
+    option(head(Head0), Options, _),
+    option(body(Body), Options, _),
     (   module_property(Module, file(File))
     ->  true
     ;   strip_module(user:Head0, Module, _)
@@ -154,6 +163,9 @@ unify_term(X1, X2) :-
 unify_term(X, Y) :-
     float(X), float(Y),
     !.
+unify_term(X, '$BLOB'(_)) :-
+    blob(X, _),
+    \+ atom(X).
 unify_term(X, Y) :-
     string(X),
     is_list(Y),
@@ -196,7 +208,7 @@ read_term_at_line(File, Line, Module, Clause, TermPos, VarNames) :-
         '$pop_input_context').
 
 read_term_at_line_2(File, Line, Module, Clause, TermPos, VarNames) :-
-    catch(try_open_source(File, In), _, fail),
+    catch(try_open_source(File, In), error(_,_), fail),
     set_stream(In, newline(detect)),
     call_cleanup(
         read_source_term_at_location(
@@ -219,6 +231,8 @@ read_term_at_line_2(File, Line, Module, Clause, TermPos, VarNames) :-
 %     prolog_clause:open_source(File, Stream)
 %     ==
 
+:- public try_open_source/2.            % used by library(prolog_breakpoints).
+
 try_open_source(File, In) :-
     open_source(File, In),
     !.
@@ -290,7 +304,14 @@ find_varname(Var, [_|T], Name) :-
 %           more complex source-translations,  falling   back  to  a
 %           heuristic method locating as much as possible.
 
-unify_clause(Read, Read, _, TermPos, TermPos) :- !.
+unify_clause(Read, _, _, _, _) :-
+    var(Read),
+    !,
+    fail.
+unify_clause(Read, Decompiled, _, TermPos, TermPos) :-
+    Read =@= Decompiled,
+    !,
+    Read = Decompiled.
                                         % XPCE send-methods
 unify_clause(Read, Decompiled, Module, TermPos0, TermPos) :-
     unify_clause_hook(Read, Decompiled, Module, TermPos0, TermPos),
@@ -323,20 +344,41 @@ unify_clause((Head :- Read),
                              ]).
                                         % DCG rules
 unify_clause(Read, Compiled1, Module, TermPos0, TermPos) :-
-    Read = (_ --> List, _),
-    is_list(List),
+    Read = (_ --> Terminal, _),
+    is_list(Terminal),
     ci_expand(Read, Compiled2, Module, TermPos0, TermPos1),
     Compiled2 = (DH :- _),
     functor(DH, _, Arity),
     DArg is Arity - 1,
+    append(Terminal, _Tail, List),
     arg(DArg, DH, List),
-    nonvar(List),
     TermPos1 = term_position(F,T,FF,FT,[ HP,
                                          term_position(_,_,_,_,[_,BP])
                                        ]),
     !,
     TermPos2 = term_position(F,T,FF,FT,[ HP, BP ]),
     match_module(Compiled2, Compiled1, Module, TermPos2, TermPos).
+unify_clause((Head,Cond => Body), Compiled1, Module,
+             term_position(F,T,FF,FT,
+                           [ term_position(_,_,_,_,[HP,CP]),
+                             BP
+                           ]),
+             TermPos) :-
+    !,
+    TermPos1 = term_position(F,T,FF,FT,
+                             [ HP,
+                               term_position(_,_,_,_,
+                                             [ CP,
+                                               term_position(_,_,_,_,
+                                                             [ FF-FT,
+                                                               BP
+                                                             ])
+                                             ])
+                             ]),
+    unify_clause((Head :- Cond, !, Body), Compiled1, Module, TermPos1, TermPos).
+unify_clause((Head => Body), Compiled1, Module, TermPos0, TermPos) :-
+    !,
+    unify_clause(Head :- Body, Compiled1, Module, TermPos0, TermPos).
                                         % general term-expansion
 unify_clause(Read, Compiled1, Module, TermPos0, TermPos) :-
     ci_expand(Read, Compiled2, Module, TermPos0, TermPos1),
@@ -453,12 +495,28 @@ ubody(B, C, _, P, P) :-
 ubody(X0, X, M, parentheses_term_position(_, _, P0), P) :-
     !,
     ubody(X0, X, M, P0, P).
-ubody(X, call(X), _,                    % X = call(X)
+ubody(X, Y, _,                    % X = call(X)
       Pos,
       term_position(From, To, From, To, [Pos])) :-
+    nonvar(Y),
+    Y = call(X),
     !,
     arg(1, Pos, From),
     arg(2, Pos, To).
+ubody(A, B, _, P1, P2) :-
+    nonvar(A), A = (_=_),
+    nonvar(B), B = (LB=RB),
+    A =@= (RB=LB),
+    !,
+    P1 = term_position(F,T, FF,FT, [PL,PR]),
+    P2 = term_position(F,T, FF,FT, [PR,PL]).
+ubody(A, B, _, P1, P2) :-
+    nonvar(A), A = (_==_),
+    nonvar(B), B = (LB==RB),
+    A =@= (RB==LB),
+    !,
+    P1 = term_position(F,T, FF,FT, [PL,PR]),
+    P2 = term_position(F,T, FF,FT, [PR,PL]).
 ubody(B, D, _, term_position(_,_,_,_,[_,RP]), TPOut) :-
     nonvar(B), B = M:R,
     ubody(R, D, M, RP, TPOut).
@@ -499,7 +557,9 @@ ubody(X0, X, M,
 ubody(X0, X, M,
       term_position(F,T,FF,TT,PA0),
       term_position(F,T,FF,TT,PA)) :-
-    expand_goal(X0, X, M, PA0, PA).
+    expand_goal(X0, X1, M, PA0, PA),
+    X1 =@= X,
+    X1 = X.
 
                                         % 5.7.X optimizations
 ubody(_=_, true, _,                     % singleton = Any
@@ -708,9 +768,10 @@ the clause as defined in init.pl
 expand_goal(G, call(G), _, P, term_position(0,0,0,0,[P])) :-
     var(G),
     !.
-expand_goal(G, G, _, P, P) :-
+expand_goal(G, G1, _, P, P) :-
     var(G),
-    !.
+    !,
+    G1 = G.
 expand_goal(M0, M, Module, P0, P) :-
     meta(Module, M0, S),
     !,
@@ -755,21 +816,21 @@ goal_expansion(send_super(R, Msg), send_class(R, _, Msg), P, P).
 goal_expansion(get_super(R, Msg, V), get_class(R, _, Msg, V), P, P).
 goal_expansion(SendSuperN, send_class(R, _, Msg), P, P) :-
     compound(SendSuperN),
-    SendSuperN =.. [send_super, R, Sel | Args],
+    compound_name_arguments(SendSuperN, send_super, [R,Sel|Args]),
     Msg =.. [Sel|Args].
 goal_expansion(SendN, send(R, Msg), P, P) :-
     compound(SendN),
-    SendN =.. [send, R, Sel | Args],
+    compound_name_arguments(SendN, send, [R,Sel|Args]),
     atom(Sel), Args \== [],
     Msg =.. [Sel|Args].
 goal_expansion(GetSuperN, get_class(R, _, Msg, Answer), P, P) :-
     compound(GetSuperN),
-    GetSuperN =.. [get_super, R, Sel | AllArgs],
+    compound_name_arguments(GetSuperN, get_super, [R,Sel|AllArgs]),
     append(Args, [Answer], AllArgs),
     Msg =.. [Sel|Args].
 goal_expansion(GetN, get(R, Msg, Answer), P, P) :-
     compound(GetN),
-    GetN =.. [get, R, Sel | AllArgs],
+    compound_name_arguments(GetN, get, [R,Sel|AllArgs]),
     append(Args, [Answer], AllArgs),
     atom(Sel), Args \== [],
     Msg =.. [Sel|Args].
